package server

import (
	"net"
	"net/http"

	"github.com/gorilla/context"

	"github.com/getlantern/golog"
	"github.com/getlantern/tlsdefaults"

	"github.com/getlantern/http-proxy/listeners"
)

var (
	testingLocal = false
	log          = golog.LoggerFor("server")
)

type listenerGenerator func(net.Listener) net.Listener

type Server struct {
	httpServer         http.Server
	listenerGenerators []listenerGenerator
}

func NewServer(handler http.Handler) *Server {
	cb := NewConnBag()

	proxy := http.HandlerFunc(
		func(w http.ResponseWriter, req *http.Request) {
			c := cb.Withdraw(req.RemoteAddr)
			context.Set(req, "conn", c)
			handler.ServeHTTP(w, req)
			context.Clear(req)
		})

	server := &Server{
		httpServer: http.Server{Handler: proxy,
			ConnState: func(c net.Conn, state http.ConnState) {
				wconn, ok := c.(listeners.WrapConn)
				if !ok {
					panic("Should be of type WrapConn")
				}

				wconn.OnState(state)

				switch state {
				case http.StateActive:
					cb.Put(wconn)
				case http.StateClosed:
					// When go server encounters abnormal request, it
					// will transit to StateClosed directly without
					// the handler being invoked, hence the connection
					// will not be withdrawed. Purge it in such case.
					cb.Purge(c.RemoteAddr().String())
				}
			},
			ErrorLog: log.AsStdLogger(),
		},
	}

	return server
}

func (s *Server) AddListenerWrappers(listenerGens ...listenerGenerator) {
	for _, g := range listenerGens {
		s.listenerGenerators = append(s.listenerGenerators, g)
	}
}

func (s *Server) ListenAndServeHTTP(addr string, readyCb func(addr string)) error {
	listener, err := net.Listen("tcp", addr)
	if err != nil {
		return err
	}
	log.Debugf("Listen http on %s", addr)
	return s.Serve(listener, readyCb)
}

func (s *Server) ListenAndServeHTTPS(addr, keyfile, certfile string, readyCb func(addr string)) error {
	listener, err := tlsdefaults.Listen(addr, keyfile, certfile)
	if err != nil {
		return err
	}
	log.Debugf("Listen https on %s", addr)
	return s.Serve(listener, readyCb)
}

func (s *Server) Serve(listener net.Listener, readyCb func(addr string)) error {
	l := listeners.NewDefaultListener(listener)

	for _, wrap := range s.listenerGenerators {
		l = wrap(l)
	}

	if readyCb != nil {
		readyCb(l.Addr().String())
	}

	return s.httpServer.Serve(l)
}
